import homepage from '@routes/index.html';
import { drizzleDB } from '@server/db';
import { entry } from '@server/db/schema';
import favicon from '@static/favicon.png';
import robotsTxt from '@static/robots.txt';
import sitemapTxt from '@static/sitemap.txt';
import { env } from 'bun';
import { desc, eq, sql } from 'drizzle-orm';

const development = env.NODE_ENV !== 'production';

const faviconInit = { headers: new Headers({ 'Content-Type': 'image/png' }) };

function getEtag(data: string) {
  const sha = Bun.SHA256.hash(data);
  return Buffer.from(sha.buffer).toString('base64url');
}

function etagResponse(etag: string, cacheControl?: string) {
  const headers = new Headers();
  headers.set('ETag', etag);
  if (cacheControl) headers.set('Cache-Control', cacheControl);
  return new Response('', { status: 304, headers });
}

let entriesETag = '';
let entriesNotModified = false;

const fetchEntries = drizzleDB
  .select({
    id: entry.id,
    name: entry.name,
    href: entry.href,
    finished: entry.finished,
    updated_at: entry.updatedAt,
  })
  .from(entry)
  .orderBy(desc(entry.updatedAt))
  .prepare();

const entryExists = drizzleDB
  .select({ id: entry.id })
  .from(entry)
  .where(eq(entry.name, sql.placeholder('name')))
  .prepare();

const insertEntry = drizzleDB
  .insert(entry)
  .values({
    name: sql.placeholder('name'),
    href: sql.placeholder('href'),
  })
  .prepare();

const updateEntry = drizzleDB
  .update(entry)
  .set({
    name: sql.placeholder('name') as any,
    href: sql.placeholder('href') as any,
  })
  .where(eq(entry.id, sql.placeholder('id')))
  .prepare();

const fetchEntryById = drizzleDB
  .select()
  .from(entry)
  .where(eq(entry.id, sql.placeholder('id')))
  .prepare();

const deleteEntry = drizzleDB
  .delete(entry)
  .where(eq(entry.id, sql.placeholder('id')))
  .prepare();

const updateEntryFinished = drizzleDB
  .update(entry)
  .set({ finished: sql.placeholder('finished') as any })
  .where(eq(entry.id, sql.placeholder('id')))
  .prepare();

Bun.serve({
  routes: {
    '/': homepage,
    '/robots.txt': new Response(robotsTxt, {
      headers: { 'Content-Type': 'text/plain' },
    }),
    '/sitemap.txt': new Response(sitemapTxt, {
      headers: { 'Content-Type': 'text/plain' },
    }),
    '/favicon.ico': new Response(favicon, faviconInit),
    '/health': new Response('OK'),
    '/api/entries': {
      async GET(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        if (entriesNotModified && req.headers.get('if-none-match') === entriesETag) {
          return etagResponse(entriesETag);
        }

        entriesNotModified = true;
        const entries = await fetchEntries.execute();

        const body = JSON.stringify(entries);
        entriesETag = getEtag(body);

        return new Response(body, {
          headers: { 'Content-Type': 'application/json', ETag: entriesETag },
        });
      },
      async PUT(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        const body = await getEntryFromReq(req);
        if (!body) return new Response('Invalid data', { status: 400 });

        const result = await entryExists.execute({ name: body.name });

        let created = true;

        if (result.length === 0) {
          await insertEntry.execute(body);
        } else if (result.length === 1) {
          const row = result[0] as NonNullable<(typeof result)[number]>;
          created = false;
          await updateEntry.execute({
            id: row.id,
            name: body.name,
            href: body.href,
          });
        } else {
          return new Response('Invalid data, multiple matches?', { status: 400 });
        }

        entriesNotModified = false;
        return new Response(JSON.stringify({ created }), {
          headers: { 'Content-Type': 'application/json' },
        });
      },
    },
    '/api/entries/:id': {
      async GET(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        const id = Number.parseInt(req.params.id, 10);
        if (Number.isNaN(id)) return new Response('Invalid id', { status: 400 });

        const result = await fetchEntryById.execute({ id });

        return new Response(JSON.stringify(result), {
          headers: { 'Content-Type': 'application/json' },
        });
      },
      async PUT(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        const id = Number.parseInt(req.params.id, 10);
        if (Number.isNaN(id)) return new Response('Invalid id', { status: 400 });

        const body = await getEntryFromReq(req);
        if (!body) return new Response('Invalid data', { status: 400 });

        await updateEntry.execute({
          id,
          name: body.name,
          href: body.href,
        });

        entriesNotModified = false;
        return new Response(JSON.stringify({ created: false }), {
          headers: { 'Content-Type': 'application/json' },
        });
      },
      async DELETE(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        const id = Number.parseInt(req.params.id, 10);
        if (Number.isNaN(id)) return new Response('Invalid id', { status: 400 });

        await deleteEntry.execute({ id });

        entriesNotModified = false;
        return new Response('OK');
      },
    },
    '/api/entries/:id/check/:finished': {
      async PUT(req) {
        if (!isAuthenticated(req)) return unauthorizedResp();

        const id = Number.parseInt(req.params.id, 10);
        if (Number.isNaN(id)) return new Response('Invalid id', { status: 400 });
        const finished = req.params.finished === 'true';

        await updateEntryFinished.execute({ id, finished });

        entriesNotModified = false;
        return new Response('OK');
      },
    },
  },
  development,
  reusePort: true,
  port: env.PORT || 3000,
  // async fetch(req, server) {
  // 	console.log(req);
  // 	return new Response("Not found", { status: 404 });
  // },
});

console.log('Server started on port:', env.PORT ? Number.parseInt(env.PORT, 10) : 3000);

async function getEntryFromReq(req: Request) {
  const json = await req.json();
  const body = json as { name: string; href: string };
  if (!body.name || !body.href || typeof body.name !== 'string' || typeof body.href !== 'string') {
    return null;
  }

  // Trim and lowercase the name to ensure consistency and uniqueness
  body.name = body.name.trim().toLocaleLowerCase();

  return body;
}

function isAuthenticated(req: Request) {
  const auth = req.headers.get('authorization');
  if (!auth) return false;
  const [type, bearer] = auth.split(' ');
  if (type !== 'Bearer') return false;
  // Check if the token is valid
  return bearer === env.BEARER_TOKEN;
}

const unauthorizedHeaders = new Headers({
  'WWW-Authenticate': `Bearer realm='sign', error="invalid_request"`,
});

function unauthorizedResp() {
  return new Response('Unauthorized', { status: 401, headers: unauthorizedHeaders });
}
